/*								TI-89 & 92P Contra!!!
								   by James Dunne
								   
											jdunne4@bradley.edu
											Jaymz031602@netscape.net
	
		As having been one of the greatest games to hit the NES console,
	I decided that a TI-89 port of Contra would be a nice tribute.  This
	source release is the beginning of a Contra game engine.  Further
	development is planned for this game of course, but I decided to release
	the source code now, so as to let others have a shot at making their own
	game before I corrupt it too much with my own game data. ;)
	
		Objects behave like the entity-driven engine of Quake, e.g. Quake-C.
	And entity is everything moving, living, and/or thinking: a player, a bullet,
	an enemy, a trigger, etc.  All entities have the same properties available.

		To spawn an entity, simply create a variable of type (entity_t *) and
	assign it to what the spawn() function returns, like so:  (see NOTE #1)
	
		entity_t  *monster = spawn();
		
	To kill the entity, simply call the kill() function, like so:
		
		kill(monster);
		
	Easy enough!
		
		See the definition of entity_t for different properties of entities.  Of course,
	there are coordinates {x, y} and there are velocities {dx, dy}. (see NOTE #2!! IMPORTANT)
	There is a lot more stuff here than I care to explain, but if you're daring enough, you
	can figure it out eventually ;).  Most of the code is pure C, and very straight-forward.
	


NOTE #1: You should only spawn entities when they come on the screen. I'll leave
	it up to you to figure out how to do that cleverly ;).  A maximum of 24 entities
	is allowed on the screen at one time.  This can be changed of course.  Simply
	change the "#define MAX_ENTITIES	24" to something else.  The array of entities
	is allocated dynamically at the beginning of _main().  This must be done before
	any spawn(), kill(), etc functions are used.
		
	
NOTE #2: Coordinates {x, y} are stored shifted by 8 bits to the left in order
	to maintain a little bit of realism, so 1 pixel on the TI = 256 units in my
	coordinate system.  This allows for pseudo-decimal increments in position
	without having to use the incredibly slow floating point routines on the TI.
	
	
*/


#define OPTIMIZE_ROM_CALLS    // Use ROM Call Optimization
#define SAVE_SCREEN           // Save/Restore LCD Contents
#include <tigcclib.h>         // Include All Header Files
// Crashlib crash-protection, will be removed after extensive debugging!  I swear!
//#include "CrashLib.h"

#define bool	unsigned char

// Compile for TI-89
// this is the only calculator I own, so :-P
	#define USE_TI89
	short _ti89;                  // Produce .89Z File
	#define __MYTI89
	#define	__LINES__		100


/*
	// Compile for TI-92P
	short _ti92plus;              // Produce .9XZ File
	#define __MYTI92P
	#define __LINES__		128
*/

// Use my own defines for the keyboard buttons:
//  Keeps up portability between TI-89 and 92.
#ifdef __MYTI89
	#define	KEYB_UP		(1<<0)
	#define KEYB_LEFT	(1<<1)
	#define KEYB_DOWN	(1<<2)
	#define KEYB_RIGHT	(1<<3)
	
	#define KEYB_2ND	(1<<4)
	#define KEYB_DIA	(1<<6)
	
	#define KEYB_F1_ROW		5
	#define KEYB_F1			(1<<7)
	#define KEYB_F2_ROW		4
	#define KEYB_F2			(1<<7)
	
	#define KEYB_ESC_ROW	6
	#define KEYB_ESC_MASK	(1<<0)

	#define KEYB_GRAY_ROW		1
	#define KEYB_GRAY_BITUP		(1<<1)
	#define KEYB_GRAY_BITDOWN	(1<<2)

	#define	TILE_W		20
	#define	TILE_H		12
	// Status line start:
	#define	LASTLINE	96

	#define	screen_w	160
	#define screen_h	__LINES__
#endif

// For the TI-92P keyboard:
#ifdef __MYTI92P
	#define KEYB_LEFT	(1<<4)
	#define	KEYB_UP		(1<<5)
	#define KEYB_RIGHT	(1<<6)
	#define KEYB_DOWN	(1<<7)
	
	#define KEYB_2ND	(1<<0)
	#define KEYB_DIA	(1<<1)
	
	#define KEYB_F1_ROW		6
	#define KEYB_F1			(1<<4)
	#define KEYB_F2_ROW		4
	#define KEYB_F2			(1<<4)

	#define KEYB_ESC_ROW	8
	#define KEYB_ESC_MASK	(1<<6)

	// FIXME:  Change these:
	#define KEYB_GRAY_ROW		1
	#define KEYB_GRAY_BITUP		(1<<1)
	#define KEYB_GRAY_BITDOWN	(1<<2)

	#define	TILE_W		30
	#define	TILE_H		16
	// Status line start:
	#define LASTLINE	128

	#define	screen_w	240
	#define screen_h	__LINES__
#endif

unsigned char *backbufl = NULL, *backbufd = NULL;
unsigned char *sprtbufl = NULL, *sprtbufd = NULL;
unsigned char *light_plane = NULL, *dark_plane = NULL;

bool	link_play = FALSE;

//#include "rle.c"

#include "sprites.c"
#include "titlescr.h"
#include "selectorspr.h"

// Although this looks like a horrible hack, its actually better to do this
// with TIGCC.  It works nicely for me, and is divided into logical subdivisions
// NOTE:  The order of #include's DOES matter.
#include "scroll.c"
#include "entities.c"
#include "bullet.c"
#include "player.c"
#include "grunt.c"

// Returns TRUE if using 2-player link-mode, FALSE for single-player.
bool main_menu() {
	int done = FALSE;
	bool twoplayer = FALSE;
	bool scrolling = TRUE;
	int	scrollx = 0;
	int y = 0, i = 0, p = 0;
	
	scrolling = TRUE;
	scrollx = 0;
	
	GrayAdjust(0);
	
	memset(light_plane, 0xFF, 3840);
	memset(dark_plane, 0xFF, 3840);

	while (!done) {
		// Frame rate limiter:
		if (!scrolling) {
			OSFreeTimer(6);
			OSRegisterTimer(6, 1);
		};

		if (scrolling && (y++ >= 100)) {
			// Scroll the title screen in from the right:
			ScrollLeft((unsigned short *)light_plane, __LINES__);
			ScrollLeft((unsigned short *)dark_plane, __LINES__);
			// Add a new line on the right side of the screen:
		#ifdef USE_TI89
			p = 19;
		#else
			p = 29;
		#endif
			i = (scrollx >> 3);
			for (y=0;y<__LINES__;y++) {
				((unsigned char *)light_plane)[p] &= ~1;
				((unsigned char *)dark_plane)[p] &= ~1;
				((unsigned char *)light_plane)[p] |= (titlescr_l[i] & (1 << (7 - (scrollx & 7)))) >> (7 - (scrollx & 7));
				((unsigned char *)dark_plane)[p] |= (titlescr_d[i] & (1 << (7 - (scrollx & 7)))) >> (7 - (scrollx & 7));
				p += 30;
				i += 30;
			};
			if (scrollx++ >= screen_w) {
				scrolling = FALSE;
				memcpy(light_plane, titlescr_l, 3840);
				memcpy(dark_plane, titlescr_d, 3840);

				// Draw one-player selector
				PortSet(light_plane, 239, 139);
				ScrRectFill(&(SCR_RECT){{4,78,19,87}}, &(SCR_RECT){{0,0,160,100}}, A_REVERSE);
				PortSet(dark_plane, 239, 139);
				ScrRectFill(&(SCR_RECT){{4,78,19,87}}, &(SCR_RECT){{0,0,160,100}}, A_REVERSE);
				
				Sprite16(4, 78, 10, selector_l, light_plane, SPRT_OR);
				Sprite16(4, 78, 10, selector_d, dark_plane, SPRT_OR);
			}
			y = 0;
		};
		
		// Adjust the grayscale:
		if (_rowread(~(1<<KEYB_GRAY_ROW)) & KEYB_GRAY_BITUP)
			GrayAdjust(grayadjust--);
		if (_rowread(~(1<<KEYB_GRAY_ROW)) & KEYB_GRAY_BITDOWN)
			GrayAdjust(grayadjust++);

		// F1 = 1 player
		if ((_rowread(~(1<<KEYB_F1_ROW)) & KEYB_F1) == KEYB_F1) {
			scrollx = screen_w - 1;
			
			PortSet(light_plane, 239, 139);
			ScrRectFill(&(SCR_RECT){{4,78,19,87}}, &(SCR_RECT){{0,0,160,100}}, A_REVERSE);
			PortSet(dark_plane, 239, 139);
			ScrRectFill(&(SCR_RECT){{4,78,19,87}}, &(SCR_RECT){{0,0,160,100}}, A_REVERSE);
			
			Sprite16(4, 78, 10, selector_l, light_plane, SPRT_OR);
			Sprite16(4, 78, 10, selector_d, dark_plane, SPRT_OR);
			
			PortSet(light_plane, 239, 139);
			ScrRectFill(&(SCR_RECT){{4,90,19,100}}, &(SCR_RECT){{0,0,160,100}}, A_NORMAL);
			PortSet(dark_plane, 239, 139);
			ScrRectFill(&(SCR_RECT){{4,90,19,100}}, &(SCR_RECT){{0,0,160,100}}, A_NORMAL);

			twoplayer = FALSE;
		};
		
		// F2 = 2 player
		if ((_rowread(~(1<<KEYB_F2_ROW)) & KEYB_F2) == KEYB_F2) {
			scrollx = screen_w - 1;
			
			PortSet(light_plane, 239, 139);
			ScrRectFill(&(SCR_RECT){{4,90,19,99}}, &(SCR_RECT){{0,0,160,100}}, A_REVERSE);
			PortSet(dark_plane, 239, 139);
			ScrRectFill(&(SCR_RECT){{4,90,19,99}}, &(SCR_RECT){{0,0,160,100}}, A_REVERSE);
			
			Sprite16(4, 90, 10, selector_l, light_plane, SPRT_OR);
			Sprite16(4, 90, 10, selector_d, dark_plane, SPRT_OR);
			
			PortSet(light_plane, 239, 139);
			ScrRectFill(&(SCR_RECT){{4,78,19,87}}, &(SCR_RECT){{0,0,160,100}}, A_NORMAL);
			PortSet(dark_plane, 239, 139);
			ScrRectFill(&(SCR_RECT){{4,78,19,87}}, &(SCR_RECT){{0,0,160,100}}, A_NORMAL);

			twoplayer = TRUE;
		};

		// Frame rate limiter:
		if (!scrolling) while (!OSTimerExpired(6)) ;

		// 2ND accepts choice:
		if ((_rowread(~(1<<0)) & KEYB_2ND) == KEYB_2ND) break;
		// ESC quits at any time:
		if ((_rowread(~(1<<KEYB_ESC_ROW)) & KEYB_ESC_MASK) == KEYB_ESC_MASK) break;
	};
	
	while (_rowread(~(1<<0)));

	return twoplayer;
};

// Main Function
void _main(void) {
	horiz_scroll = TRUE;

	int i, j;
	int done = 0;
	
	time = 0;
	
	long	oldscrn_x = 0, oldscrn_y = 0;
	
	// Initialize all memory before going into "graphics mode":
	
	// Maybe in a future release, use the gray.h GrayDBuf...() functions instead of
	// this old way:  (might save some memory in the process as well??)
	backbufl = (unsigned char *)safe_malloc(LCD_SIZE);
	if (backbufl == NULL) return;
	backbufd = (unsigned char *)safe_malloc(LCD_SIZE);
	if (backbufd == NULL) return;
	sprtbufl = (unsigned char *)safe_malloc(LCD_SIZE);
	if (sprtbufl == NULL) return;
	sprtbufd = (unsigned char *)safe_malloc(LCD_SIZE);
	if (sprtbufd == NULL) return;
	
	ClrScr();

	INT_HANDLER save_int_1 = GetIntVec (AUTO_INT_1);
	SetIntVec (AUTO_INT_1, DUMMY_HANDLER);   // redirect auto-int 1 to "nothing"
	// enable grayscale
	GrayOn();									//turn grayscale on

	ScrRectFill(&(SCR_RECT){{0,0,160,100}}, &(SCR_RECT){{0,0,160,100}}, A_REVERSE);		//turns screen white
	
	light_plane = GetPlane(LIGHT_PLANE);
	dark_plane = GetPlane(DARK_PLANE);
	
	// Main-menu: (returns TRUE if using two-player cable-link mode)
	link_play = main_menu();

	GrayAdjust(grayadjust = -20);

	// Initialization of game-data:
	entity_t	*curr;

	// Must be called before any spawn(), kill(), etc...
	entities = (entity_t **)calloc(MAX_ENTITIES, sizeof(entity_t *));
	
	map	= (unsigned char *)calloc(TILE_W * TILE_H, sizeof(unsigned char));
	if (map == NULL) return;
	unsigned char *mapi = map;

	player = player_spawn();
	player->x = 256 * 8;
	player->y = 256 * 8;
	unsigned char playerindex = player->index;

	// Messages and other immediate drawing commands are re-routed to the dark-sprite buffer
	// and will only be displayed when the sprite buffer is swapped to the screen
	PortSet (sprtbufd, 239, 127);

	// Draw initial background:
	memset(backbufl, 0b00000000, LCD_SIZE);
	memset(backbufd, 0b00000000, LCD_SIZE);

	randomize();

	// Background screen-scrolling coordinates defining the upper-left corner of
	// the view screen.
	screen_x = 0;
	screen_y = 0;
	
	// Draw first background layout:
	int x, y, s = 0;
	
	// The global (unsigned char *)map variable contains a TILE_W x TILE_H tile-map
	// of the CURRENT VISIBLE SCREEN.  Each tile # represents the corresponding
	// 8x8 tile index in the backsprites[] array defined in 'sprites.c'
	mapi = map;
	for (y=0; y<TILE_H; y++) {
		for (x=0; x<TILE_W; x++) {
			// Simple one-screen test map used to create & debug collision detection and scrolling.
			if ((x == 0) || (y == 0) || (y == TILE_H-1)) *mapi = 1 | 0x80; else *mapi = 0;
			if (((x % 2) == 1) && (y == (TILE_H - 2))) *mapi = 1 | 0x80;
			s = *mapi++ & 0x7f;
			Sprite8(x << 3, y << 3, 8, backsprites[(s << 1) + 0], backbufl, A_OR);
			Sprite8(x << 3, y << 3, 8, backsprites[(s << 1) + 1], backbufd, A_OR);
		};
	};

	// TAKE NOTE:
		// Define the map-file structure early on to be expandable and backward-compatible.
		// Perhaps a chunk-based structure where only known chunks are handled, with the
		// support for extra and future added information.
	
		// I would recommend an RLE-based compression of the map-files' MAP chunk data. In
		// fact, an RLE implementation has been incorporated into the source, 'rle.c', but is
		// currently not being used.
	
		// For being a mainly TI-89/92P specific game, I would like to create two versions of
		// the code; one being ANSI-standard, and the other being TI-specific -- to increase
		// speed on the TI's and also to maintain portability.
	
		// In order to become TI-specific, I would recommend a rewrite of the RLE algorithm's file
		// access code to incorporate direct VAT access, rather than TIGCCLIB's ANSI standard compliance
		// of using fopen, fread, fclose, etc functions.
	
	oldscrn_x = screen_x;
	oldscrn_y = screen_y;
	
	// Main loop:
	while (entities[playerindex] != NULL) {
		// Frame rate limiter:
		OSFreeTimer(6);
		OSRegisterTimer(6, 1);

		// Adjust the grayscale:
		if (_rowread(~(1<<KEYB_GRAY_ROW)) & KEYB_GRAY_BITUP)
			GrayAdjust(grayadjust--);
		if (_rowread(~(1<<KEYB_GRAY_ROW)) & KEYB_GRAY_BITDOWN)
			GrayAdjust(grayadjust++);
		
		// Maybe implement switchable scrolling methods i.e. original Contra
		// one level is vertical scrolling only, another is horizontal scrolling only.
		// vertical scrolling can be up or down, horizontal only right
		
		// Default is side-scrolling, in fact the map handling code will
		// be biased towards loading the rightmost row in order to scroll
		// the map correctly:
		
		// Maybe we should implement a left-scrolling algorithm as well for some
		// levels, so we don't get the original Contra's feeling of heading EAST all
		// the time.
		
		/*if (horiz_scroll)*/ {
			// Scroll the background left one line:
			if (screen_x > oldscrn_x) {
				for (i=oldscrn_x; i<screen_x; i++) {
					ScrollLeft((unsigned short *)backbufl, __LINES__);
					ScrollLeft((unsigned short *)backbufd, __LINES__);
				};
				// Determing starting row to draw from:
				// Using two pointers for this is assumedly faster access due
				// to not having to calculate the index for each column, since
				// we already know the width of the visible map:
				mapi = &( map[ (screen_x >> 3) % TILE_W ] );
				for (y=0; y<TILE_H; y++) {
					s = *mapi & 0x7F;
					Sprite8((TILE_W << 3) - (screen_x % 8), (y << 3), 8, backsprites[(s << 1) + 0], backbufl, A_OR);
					Sprite8((TILE_W << 3) - (screen_x % 8), (y << 3), 8, backsprites[(s << 1) + 1], backbufd, A_OR);
					// Next column:
					mapi += TILE_W;
				};
			}
		} /*else {
			// Not tested, but theoretically should be correct:
			// Note, if you want up AND downward scrolling levels,
			// you'll have to have a global control flag which
			// specifies which direction to scroll and change the
			// code below accordingly:
			if (screen_x > oldscrn_x) {
				for (i=oldscrn_y; i<screen_y; i++) {
					ScrollDown((unsigned short *)backbufl, __LINES__);
					ScrollDown((unsigned short *)backbufd, __LINES__);
				};
				mapi = map[TILE_W - 1];
				for (x=0; x<TILE_H; y++) {
					s = *mapi;
					Sprite8((x << 3) - (screen_x % 8), 0 - (screen_y % 8), 8, backsprites[(s << 1) + 0], backbufl, A_OR);
					Sprite8((x << 3) - (screen_x % 8), 0 - (screen_y % 8), 8, backsprites[(s << 1) + 1], backbufd, A_OR);
					mapi += TILE_W;
				};
			}
		}*/
		
		// pause button = 2nd: (horrible hack for debugging purposes)
		//while (_rowread(~(1<<0)) & KEYB_2ND);
		
		// The background is copied into the sprite-buffer:
		memcpy(sprtbufl, backbufl, LCD_SIZE);
		memcpy(sprtbufd, backbufd, LCD_SIZE);

		// Prevent unsigned int from overflowing
		if (player->x < screen_x + 256) player->x = screen_x + 256;
		if (player->y < screen_y + 256) player->y = screen_y + 256;

		oldscrn_x = screen_x;
		oldscrn_y = screen_y;

		if (time % 70 == rand() % 70) {
			spawn_grunt();
		};
		
		// Handle thinking, collisions, and rendering for every entity:
		for (i = 0; i < MAX_ENTITIES; i++) if (entities[i] != NULL) {
			curr = entities[i];
			
			// Check for collisions among entities:
			for (j = 0; j < MAX_ENTITIES; j++) if ((entities[j] != NULL) && (j != i)) {
				if (entities[i] == NULL) break;
				if (collide(entities[i], entities[j])) {
					if (entities[i]->touch) entities[i]->touch(entities[i], entities[j]);
					if (entities[i] == NULL) break;
					if (entities[j]->touch) entities[j]->touch(entities[j], entities[i]);
				}
			};

			if (entities[i] == NULL) continue;
			
			// Check for ground & wall collisions:
			curr->colflags = check_map(curr);
			if ((curr->colflags & BLOCKED_GROUND) == BLOCKED_GROUND) {
				curr->y = (curr->y >> 11) << 11;
				curr->dy = 0;
				curr->flags |= FLAG_ONGROUND;
			} else {
				curr->flags &= ~FLAG_ONGROUND;
			};
			
			// Think for the entity (if it's time):
			if ((curr->think != NULL) && (time == curr->next_think))	curr->think(curr);
			
			// We have to check if this still exists, because the entity
			// could have removed itself via kill(self):
			// We can NOT check if curr == NULL because curr (as a pointer) is not passed
			// by reference, and therefore will not reflect the changes to its master
			// pointer in entites[], unless we use curr as a double-pointer?? :-D
			if (entities[i] == NULL) continue;

			// Check if we should bother to draw first:
			if (curr->render != NULL)
				// Is it within screen bounds?
				if ((curr->x >> 8 >= screen_x) && (curr->x >> 8 <= screen_x + screen_w) && 
					(curr->y >> 8 >= screen_y) && (curr->y >> 8 <= screen_y + screen_h)) 
					curr->render(curr);
			
			// Apply the deltas: (velocity)
			curr->x += curr->dx;
			curr->y += curr->dy;
			
			// Off the edge of the screen?  Kill it!
			if (curr != player) {
				if ((curr->x >> 8) < screen_x) { kill(curr); continue; }
				if ((curr->x >> 8) > 4 + screen_x + screen_w) { kill(curr); continue; }
			};
			
			if (entities[i] == NULL) continue;
			
			// Handle scrolling the screen for CLASS_PLAYER's:
			if (curr->class == CLASS_PLAYER) {
				if ((curr->x >> 8) - screen_x > 80) screen_x = ((curr->x >> 8) - 80);
				if (screen_x < 0) screen_x = 0;
			}
			
			// Gravity (only when not ONGROUND and not NOGRAVITY):
			if (!(curr->flags & FLAG_ONGROUND) && !(curr->flags & FLAG_NOGRAVITY))
				curr->dy += 220;
			
			// Friction (only when not NOFRICTION):
			if (!(curr->flags & FLAG_NOFRICTION)) {
				// Since we have scaled everything up by 256, we can't multiply
				// by 0.1 or anything small like that, because they're not decimals,
				// so, alas, divides will have to suffice.  But wait!  Right-Shifts
				// work FASTER!!!
				if (curr->dx < -32) curr->dx = -(abs(curr->dx) >> 1);
				else if (curr->dx > 32) curr->dx >>= 1;
				else curr->dx = 0;
			};
			
			// Ground check:
			// Horrible hack, should be tile-based checking, and will be soon:
			if (((curr->y >> 8) - screen_y) > (LASTLINE - curr->height)) {
				curr->y = ((LASTLINE - curr->height) << 8) + screen_y;
				curr->dy = 0;
				curr->flags |= FLAG_ONGROUND;
			};

			curr->oldx = curr->x;
			curr->oldy = curr->y;
		};

		// The sprites & background are rendered in sprtbufl & d.
		// Copy them to the LCD
		memcpy(light_plane, sprtbufl, LCD_SIZE);
		memcpy(dark_plane, sprtbufd, LCD_SIZE);

		// Frame rate limiter:
		while (!OSTimerExpired(6)) ;
		
		time++;
		
		// ESC quits at any time:
		if ((_rowread(~(1<<KEYB_ESC_ROW)) & KEYB_ESC_MASK) == KEYB_ESC_MASK) break;
	};
	
	// Remove all entities:
	entity_t	*next;
	
	for (i = 0; i < MAX_ENTITIES; i++) {
		kill(entities[i]);
		entities[i] = NULL;
	};
	
	free(map);

	free(entities);
	
	free(sprtbufl);
	free(sprtbufd);

	free(backbufl);
	free(backbufd);
	
	GrayAdjust(0);
	
	GrayOff();
	
	// disable grayscale
	SetIntVec (AUTO_INT_1, save_int_1);
	
	ClrScr();
	return;
};
