void draw_player(struct entity_t *self) {
	// This is the general function to simply render a player-looking
	// sprite:
	if (self == NULL) return;
	int x = (self->x >> 8) - screen_x;
	int y = (self->y >> 8) - screen_y;

	if (self->flags & FLAG_LIEDOWN) {
		if (self->flags & FLAG_FACERIGHT) {
			Sprite8(x, y + 8, 8, sprites_player[(0x03 << 2) + 0], sprtbufl, A_OR);
			Sprite8(x, y + 8, 8, sprites_player[(0x03 << 2) + 1], sprtbufd, A_OR);
			Sprite8(x + 8, y + 8, 8, sprites_player[(0x02 << 2) + 0], sprtbufl, A_OR);
			Sprite8(x + 8, y + 8, 8, sprites_player[(0x02 << 2) + 1], sprtbufd, A_OR);
		} else {
			Sprite8(x, y + 8, 8, sprites_player[(0x03 << 2) + 2], sprtbufl, A_OR);
			Sprite8(x, y + 8, 8, sprites_player[(0x03 << 2) + 3], sprtbufd, A_OR);
			Sprite8(x - 8, y + 8, 8, sprites_player[(0x02 << 2) + 2], sprtbufl, A_OR);
			Sprite8(x - 8, y + 8, 8, sprites_player[(0x02 << 2) + 3], sprtbufd, A_OR);
		};
	} else {
		if (self->flags & FLAG_FACERIGHT) {
			Sprite8(x, y, 8, sprites_player[(0x00 << 2) + 0], sprtbufl, A_OR);
			Sprite8(x, y, 8, sprites_player[(0x00 << 2) + 1], sprtbufd, A_OR);
			Sprite8(x, y + 8, 8, sprites_player[(0x01 << 2) + 0], sprtbufl, A_OR);
			Sprite8(x, y + 8, 8, sprites_player[(0x01 << 2) + 1], sprtbufd, A_OR);
		} else {
			Sprite8(x, y, 8, sprites_player[(0x00 << 2) + 2], sprtbufl, A_OR);
			Sprite8(x, y, 8, sprites_player[(0x00 << 2) + 3], sprtbufd, A_OR);
			Sprite8(x, y + 8, 8, sprites_player[(0x01 << 2) + 2], sprtbufl, A_OR);
			Sprite8(x, y + 8, 8, sprites_player[(0x01 << 2) + 3], sprtbufd, A_OR);
		};		
	};
};

void render_player(struct entity_t *self) {
	int i;

	// Draw the player:
	draw_player(self);
	
	// Render health bar:
	for (i = 0; i < (self->max_health >> 5); i++) {
		if (self->health >= (i << 5)) {
			if (self->health >= (i << 5) + 16) {
				Sprite8(8 + (i << 3), 8, 8, sprites_healthbar[(0x02 << 2) + 0], sprtbufl, A_OR);
				Sprite8(8 + (i << 3), 8, 8, sprites_healthbar[(0x02 << 2) + 1], sprtbufd, A_OR);
			} else {
				Sprite8(8 + (i << 3), 8, 8, sprites_healthbar[(0x01 << 2) + 0], sprtbufl, A_OR);
				Sprite8(8 + (i << 3), 8, 8, sprites_healthbar[(0x01 << 2) + 1], sprtbufd, A_OR);
			}
		} else {
			Sprite8(8 + (i << 3), 8, 8, sprites_healthbar[(0x00 << 2) + 0], sprtbufl, A_OR);
			Sprite8(8 + (i << 3), 8, 8, sprites_healthbar[(0x00 << 2) + 1], sprtbufd, A_OR);
		}
	};
};


// Basic control thinking function for the player:  (handle buttons)
void player_think(struct entity_t *self) {
	unsigned char gamebuttons = _rowread(~(1<<0));
	static unsigned char lastbuttons = 0;
	
	if (self == NULL) return;
	// Handle buttons:
	if (gamebuttons & KEYB_DIA) {
		if ((self->flags & FLAG_ONGROUND) && !(self->flags & FLAG_LIEDOWN)) {
			self->dy -= 1400; self->flags &= ~FLAG_ONGROUND;
		}
	};
	if ((gamebuttons & KEYB_2ND) && (time >= self->attack_time)) {
		// Shoot:
		self->attack_time = time + 5;
		entity_t *bullet = spawn_bullet(self, 
			((self->flags & FLAG_FACERIGHT) ? 
				(((self->flags & FLAG_LIEDOWN) ? 16 : 8) << 8) : 
				-(((self->flags & FLAG_LIEDOWN) ? 16 : 8) << 8)), 
			8 << 8, self->attack_damage);
	};
	if (gamebuttons & KEYB_UP)		{ self->flags |= FLAG_LOOKUP; self->flags &= ~FLAG_LIEDOWN; } else self->flags &= ~FLAG_LOOKUP;
	if (gamebuttons & KEYB_DOWN)	{
		if (self->flags & FLAG_ONGROUND) {
			self->flags |= FLAG_LIEDOWN;
			//if (!(lastbuttons & KEYB_DOWN)) self->sprite = &(spritelayout[1]);
		}
	} else {
		self->flags &= ~FLAG_LIEDOWN;
		//if ((lastbuttons & KEYB_DOWN)) self->sprite = &(spritelayout[0]);
	}

	if (gamebuttons & KEYB_LEFT)	{
		if (!(self->flags & FLAG_LIEDOWN)) self->dx = -512;
		self->flags &= ~FLAG_FACERIGHT;
	};
	if (gamebuttons & KEYB_RIGHT)	{
		if (!(self->flags & FLAG_LIEDOWN)) self->dx = 512;
		self->flags |= FLAG_FACERIGHT;
	};

	self->next_think = time + 1;
	lastbuttons = gamebuttons;
};


void msg_render(struct entity_t *self) {
	printf_xy(58, 48, "GAME OVER");
};


void player_death(struct entity_t *self) {
	self->dy = -1210;
	self->dx = (self->flags & FLAG_FACERIGHT) ? 1712 : -1712;
	self->flags |= FLAG_LIEDOWN | FLAG_DEAD;
	self->flags &= ~FLAG_ONGROUND;
	
	self->death = NULL;		// dont call death again!
	self->think = (func_t *)sub_remove;
	self->next_think = time + 90;
	
	// Create an entity to display a message:
	entity_t *msg = spawn();
	msg->owner = self;
	msg->x = self->x; msg->y = self->y;
	msg->height = 0;
	msg->width = 0;
	msg->flags = FLAG_NOGRAVITY | FLAG_NOFRICTION;

	msg->render = (func_t *)msg_render;
	msg->think = (func_t *)sub_remove;
	msg->next_think = time + 90;
};

entity_t *player_spawn() {
		// Tie the player entity to the 1st entity in the list:
	entity_t *p = spawn();
	if (p == NULL) return;

	p->class = CLASS_PLAYER;
	// Set its rendering function appropriately:
	p->next_think = 0;
	p->think = (func_t *)player_think;
	p->death = (func_t *)player_death;
	p->render = (func_t *)render_player;

	p->width = 8; p->height = 16;
	p->x = 0;
	p->y = 0;
	
	p->health = p->max_health = 128;
	p->attack_damage = 32;			// Player's bullets do 32 damage.
	
	return p;
};
