// Thanks to the developers of CRGOUE for these routines.
// I could've written them myself, but why re-invent the wheel?

#include <stdio.h>
#include "rle.h"

unsigned int runlen=0;
unsigned char runchar=' ';

unsigned char _rleread_file(void);
unsigned char _rleread_ptr(void);

void *rlepos;
unsigned char (*_rleread)(void);


const unsigned char magic_byte = 0x5F;

//{{{
void frlewrite(void *data, size_t s, size_t n, FILE *f)
{
	register int i, j;
	
	for(i=0; i<s; i++)
		for(j=0; j<n; j++)
			frleputc( ((char*)data)[j*s+i], f, 0);
}
//}}}
//{{{
void frleread(void *data, size_t s, size_t n, FILE *f)
{
	register int i, j;
	
	rle_set_source(rle_file, f);
	
	for(i=0; i<s; i++)
		for(j=0; j<n; j++)
			((char*)data)[j*s+i] = rlegetc();
}
//}}}
//{{{
//
// Writes a run-length encoded character to a file
// Uses static data to keep track of state; the 'mode' parameter
// is used to indicate special conditions
// mode: 0 for normal write, 1 for end of write, 2 for initialize
//
void frleputc(unsigned char c, FILE *fout, unsigned char mode)
{
	if( mode==2 )
	{
		runlen=0;
		runchar=0;
		return;
	}

	if( runlen == 0 )
	{
		runlen=1;
		runchar=c;
		return;
	}
	
	if( runlen >= 0x7FFF || c != runchar || mode==1 )
	{
		if(runlen > 127)
		{
			fputc( magic_byte, fout );
			fputc( runchar, fout );
			fputc( (runlen>>8)|0x80, fout );
			fputc( runlen & 0xFF, fout );
		}
		else if(runlen>2 || runchar==magic_byte) // encode only runs that are long enough to be profitable
		{
			fputc( magic_byte, fout );
			fputc( runchar, fout );
			fputc( runlen, fout );
		}
		else if(runlen==2)
		{
			fputc( runchar, fout );
			fputc( runchar, fout );
		}
		else
		{
			fputc( runchar, fout );
		}
		runchar = c;
		runlen = 1;
	} else {
		runlen++;
	}
	if(mode==1)
	{
		runlen = 0;
		runchar = 0;
	}
}
//}}}

//{{{
//
// Reads a run-length encoded character from a file
// Uses static data to keep track of state
//
unsigned char rlegetc(void)
{
	unsigned char in;
	
	if( !runlen ) {
		in = _rleread();
		
		if(in!=magic_byte)
			return in;
		
		runchar = _rleread();
		runlen = _rleread();
		if(runlen & 0x80) {
			runlen &= ~0x80;
			runlen <<= 8;
			runlen |= _rleread();
		}
	}
	runlen--;
	return runchar;
}
//}}}

//{{{
void rle_set_source(rle_source_type t, void *s)
{
	rlepos = s;
	
	switch(t)
	{
		case rle_file:
			_rleread = _rleread_file;
			break;
		case rle_pointer:
			_rleread = _rleread_ptr;
			break;
	}
}
//}}}
//{{{
unsigned char _rleread_file(void)
{
	return fgetc( (FILE*)rlepos );
}
//}}}
//{{{
unsigned char _rleread_ptr(void)
{
	return *( ((unsigned char*)rlepos) ++);
}
//}}}

