
// Create a mirror image of an 8-bit pattern:
#define FLIP(x) (unsigned char)( (((unsigned char)x & (1<<7)) >> 7) | \
				(((unsigned char)x & (1<<6)) >> 5) | \
				(((unsigned char)x & (1<<5)) >> 3) | \
				(((unsigned char)x & (1<<4)) >> 1) | \
				(((unsigned char)x & (1<<3)) << 1) | \
				(((unsigned char)x & (1<<2)) << 3) | \
				(((unsigned char)x & (1<<1)) << 5) | \
				(((unsigned char)x & (1<<0)) << 7) )

// This macro defines the sprite's light & dark planes (grayscale) and
// creates a mask of that image for background restoration.  Then it
// also creates a mirror image version of the same two planes, and also
// creates a mask for that mirror image.
#define	DEF_SPRITE(a1, a2, b1, b2, c1, c2, d1, d2, e1, e2, f1, f2, g1, g2, h1, h2) \
		{a1, b1, c1, d1, e1, f1, g1, h1}, \
		{a2, b2, c2, d2, e2, f2, g2, h2}, \
		{FLIP(a1), FLIP(b1), FLIP(c1), FLIP(d1), FLIP(e1), FLIP(f1), FLIP(g1), FLIP(h1)}, \
		{FLIP(a2), FLIP(b2), FLIP(c2), FLIP(d2), FLIP(e2), FLIP(f2), FLIP(g2), FLIP(h2)}

// Same as above, but without mirror-image sprites:
#define	DEF_SPRITE_BG(a1, a2, b1, b2, c1, c2, d1, d2, e1, e2, f1, f2, g1, g2, h1, h2) \
		{a1, b1, c1, d1, e1, f1, g1, h1}, \
		{a2, b2, c2, d2, e2, f2, g2, h2}

/*
		I guess this sprite handling junk is in need of a little explaining. :)
	If you are not concerned with the details of how this works, then dont be ;).
	Contra supports, and depends heavily on, 4-level greyscale graphics.  To define
	a sprite, use the DEF_SPRITE() macro.  The left side is the light-plane pixel
	mask, and the right side is the dark-plane pixel mask.  I did this because I do
	all my sprite editing in the code, and this is the best way I could come up with,
	since I don't have a 4-level greyscale sprite-editor for Linux. :-D.  If you have
	a better way of doing this, then by all means, implement it.
		The reason this is macro is so handy, is because it automatically defines 4
	sprites for every 1 sprite:  the light-plane sprite, the dark-plane sprite, the
	REVERSED (mirror-image) light-plane sprite, and the REVERSED dark-plane sprite.
	Take a look at the DEF_SPRITE() macro definition above to see how this works.	
		The sprite drawing code expects sprites to be in this order, so take note of this
	before you go fiddling with the internals of it all.
*/

static unsigned char sprites_player[16][8] = {
		DEF_SPRITE( 0b00000000, 0b00000000,
					0b00111100, 0b00111100,
					0b01111110, 0b01111110,
					0b01110010, 0b01100110,
					0b01100010, 0b01000010,
					0b01100110, 0b01100110,
					0b00111100, 0b00111100,
					0b00011000, 0b00011000 ),

		DEF_SPRITE( 0b00111000, 0b00011000,
					0b00111000, 0b00111100,
					0b01111111, 0b01111111,
					0b00011000, 0b00111000,
					0b00111000, 0b00111000,
					0b00111100, 0b00111100,
					0b01101100, 0b01101100,
					0b01101100, 0b01101100 ),

		// Lying down:
		DEF_SPRITE( 0b00000000, 0b00000000,
					0b00111100, 0b00111100,
					0b01111110, 0b01111110,
					0b01110010, 0b01100110,
					0b11100010, 0b11000010,
					0b11100110, 0b11100110,
					0b00111100, 0b00111100,
					0b00000000, 0b00000000 ),

		DEF_SPRITE( 0b00000000, 0b00000000,
					0b00000000, 0b00000000,
					0b00000000, 0b00000000,
					0b00000000, 0b00000000,
					0b00011110, 0b01111001,
					0b11111111, 0b11111110,
					0b11101111, 0b10001100,
					0b00011100, 0b00011100 ),
};


static unsigned char sprites_bullet[4][8] = {
		// Bullet:
		DEF_SPRITE( 0b00000000, 0b00000000,
					0b00000000, 0b00000000,
					0b00010000, 0b00010000,
					0b00111100, 0b00111000,
					0b00011000, 0b00010000,
					0b00010000, 0b00000000,
					0b00000000, 0b00000000,
					0b00000000, 0b00000000 ),
};

static unsigned char sprites_healthbar[12][8] = {
		// Status-bar sprites:
		DEF_SPRITE(	0b00000000, 0b00000000,
					0b01100110, 0b00000000,
					0b10011001, 0b00000000,
					0b10000001, 0b00000000,
					0b10000001, 0b00000000,
					0b01000010, 0b00000000,
					0b00100100, 0b00000000,
					0b00011000, 0b00000000 ),

		DEF_SPRITE(	0b00000000, 0b00000000,
					0b01100110, 0b01100000,
					0b10011001, 0b11111000,
					0b10000001, 0b11110000,
					0b10000001, 0b11111000,
					0b01000010, 0b01110000,
					0b00100100, 0b00111000,
					0b00011000, 0b00011000 ),

		DEF_SPRITE(	0b00000000, 0b00000000,
					0b01100110, 0b01100110,
					0b10011001, 0b11111111,
					0b10000001, 0b11111111,
					0b10000001, 0b11111111,
					0b01000010, 0b01111110,
					0b00100100, 0b00111100,
					0b00011000, 0b00011000 ),
};


// Background sprites (dont have to be reversed...)
static unsigned char backsprites[4][8] = {
		DEF_SPRITE_BG(	0b00000000, 0b00000000,
						0b00000000, 0b00000000,
						0b00000000, 0b00000000,
						0b00000000, 0b00000000,
						0b00000000, 0b00000000,
						0b00000000, 0b00000000,
						0b00000000, 0b00000000,
						0b00000000, 0b00000000 ),

		DEF_SPRITE_BG(	0b11111111, 0b00000000,
						0b10000001, 0b01000010,
						0b10000001, 0b00100100,
						0b10000001, 0b00011000,
						0b10000001, 0b00011000,
						0b10000001, 0b00100100,
						0b10000001, 0b01000010,
						0b11111111, 0b00000000 ),
};

